'''ard_collection provides classes for working with stored ARD tiles

The module has 3 primary classes:

- ARDCollection: a collection of ARD tiles
- ARDTile: a representation of an ARD tile
- Acquisition: a container of ARDTiles from a given acquisition

  Typical usage example:

  c = ARDCollection('bucket/prefix')
  for tile in c.tiles:
      print tile.cell.bounds
'''
from __future__ import annotations

import warnings
import os
import re
import json
from typing import List, Iterable, Optional, Union, Any
from datetime import date, datetime
from collections import defaultdict
from posixpath import join as urljoin
from pathlib import Path
from math import floor
from functools import lru_cache

from fsspec import filesystem
from fsspec.spec import AbstractFileSystem
from maxar_canvas_grid import covers, Cell
from maxar_canvas_grid.grid import get_transform
from max_ard.exceptions import MissingDependency
from max_ard.order import Order
from pydantic.utils import to_camel
from shapely.geometry.polygon import Polygon

# Conditional imports so we can have a simpler base install

try:
    from shapely.geometry import mapping, shape, GeometryCollection
    from shapely.ops import transform
    HAS_SHAPELY = True
except ImportError:
    HAS_SHAPELY = False

try:
    import rasterio
    HAS_RASTERIO = True
except ImportError:
    HAS_RASTERIO = False

try:
    import fiona
    HAS_FIONA = True
except ImportError:
    HAS_FIONA = False

# Python < 3.7 compatibility

try:
    from functools import cached_property
except ImportError:
    from backports.cached_property import cached_property

# optimization from GBDX, not currently re-implemented
#catalog_url = 'https://geobigdata.io/catalog/v2/record/'


class ARDTile():
    ''' Represents an ARD Tile - the images and vectors from one acquistion in one grid cell

    '''


    # matching regexes
    # don't be a smartass and combine them

    # matches quadkeys
    CELL_REGEX = re.compile(r'[/-](\d{2})[-/]([0123]{12})[/-]')
    # matches regular catalog IDs
    CAT_ID_REGEX = re.compile(r'10[1-6][0AC][0-9A-F]{3}0[0-9A-F]{6}00', re.I)
    # matches WV4 catIDs (UUID v4)
    WV4_REGEX = re.compile(r'([0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}-inv)', re.I)
    # matches dates yyyy-mm-dd
    DATE_REGEX = re.compile(r'\d{4}-\d{2}-\d{2}')

    def __init__(self) -> None:
        self.date = None #: date of cell
        
        #: acquisition ID
        self.acq_id = None
        self.quadkey = None
        ''' ARD Grid quadkey of the tile's cell '''
        self.zone = None
        ''' UTM zone of the tile '''
        self.asset_paths = {}
        ''' paths to the tile's assets '''

    @cached_property
    def properties(self) -> dict:
        ''' metadata of the ARD tile '''
        return self.open_asset('item')['properties']

    @cached_property
    def clouds(self) -> GeometryCollection:
        ''' Cloud polygon geometry '''
        return self.open_asset('clouds')

    @cached_property
    def nodata(self) -> Polygon:
        ''' Extent of pixels that do not have valid data '''
        # TODO check projections
        return self.cell.geom_WGS84.difference(self.extent)

    @cached_property
    def extent(self) -> Polygon:
        ''' Extent of pixels that have valid data '''
        return self.open_asset('extent')
    
    @cached_property
    def cell(self) -> Cell:
        ''' The ARD Grid cell of the tile '''
        return Cell(self.quadkey, zone=self.zone)


    @classmethod
    def from_doc(cls, fs: AbstractFileSystem, path: str) -> ARDTile:
        ''' Create an ARDTile object from a path to any of a tile's files.

        Args:
            fs: a `filesystem` object used to read files
            path: a path to any file in a tile
        '''

        tile = cls()
        tile.fs = fs

        # zone and quadkey
        matches = ARDTile.CELL_REGEX.search(path)
        if matches is not None:
            tile.zone, tile.quadkey= matches.groups()

        # cat ID
        matches = ARDTile.CAT_ID_REGEX.search(path)
        if matches is not None:
            tile.acq_id = matches[0]
        else:
            matches = ARDTile.WV4_REGEX.search(path)

        # date
        matches = ARDTile.DATE_REGEX.search(path)
        if matches is not None:
            tile.date = matches[0]
            ''' Date of the tile '''
         
        # prefixes for opening files with gdal/ogr
        if 's3' in tile.fs.protocol:
            tile.gdal_prefix = '/vsis3/'
        else:
            tile.gdal_prefix = ''

        # rebuild the path from properties
        # might be overkill
        base_prefix = path.split(f"/{tile.zone}/{tile.quadkey}")[0]
        base_prefix = urljoin(base_prefix, tile.zone, tile.quadkey)
        if tile.date is not None:
            base_prefix = urljoin(base_prefix, tile.date)
        base_prefix = urljoin(base_prefix, tile.acq_id)
        tile.asset_paths = {
            'visual': base_prefix + '-visual.tif',
            'pan': base_prefix + '-pan.tif',
            'ms': base_prefix + '-ms.tif',
            'item': base_prefix + '.json',
            'clouds': base_prefix + '-cloud-mask.gpkg',
            'extent': base_prefix + '-data-mask.gpkg'
        }
        ''' paths to the assets '''

        tile.base_prefix = base_prefix

        return tile

    def __repr__(self):
        return f'<ARDTile of {self.acq_id} at z{self.zone}-{self.quadkey}>'
    
    def open_asset(self, name):
        ''' Opens an asset by name

            COGs returns a Rasterio DataSetReader
            Masks return Shapely geometries
            JSON returns dict '''


        asset = self.asset_paths[name]
        if asset.endswith('tif'):
            if not HAS_RASTERIO:
                raise MissingDependency('Rasterio is required to read ARD rasters')
            return rasterio.open(f'{self.fs_prefix}{asset}')

        if asset.endswith('geojson') or asset.endswith('gpkg'): 
            if not HAS_FIONA:
                raise MissingDependency('Fiona is required to access ARD masks as geometries')
            if not HAS_SHAPELY:
                raise MissingDependency('Shapely is required to access ARD masks as geometries')
            with fiona.open(f'{self.fs_prefix}{asset}') as layer:
                geom = [shape(f['geometry']) for f in layer]
                proj = layer.crs['init']
                if proj != 'epsg:4326':
                    tfm = get_transform(proj, 4326)
                    geom = [transform(tfm, f) for f in geom]
            if len(geom) == 1:
                return geom[0]
            elif len(geom) == 0:
                return None
            else:
                return GeometryCollection(geom)
            

        if asset.endswith('json'):
            with self.fs.open(asset) as f:
                return json.load(f)


    def __geo_interface__(self):
        # return cell extents or pixel extents?
        return mapping(self.extent)

class ARDCollection():
    ''' ARDCollections represent collections of S3 tiles. Currently the tiles
        can be stored in S3 or locally. 
        
        The following attributes are settable and will trigger a rescan 
        Attributes:
            path: path to S3 prefix or STAC collection
            aoi: limit to finding tiles that cover this AOI, can be shapely geometry or most textual representations
            acq_id_in: limit to finding tiles from these acquisitions
            zone_in: limit to finding tiles in these zones
            earliest_date: limit to finding tiles after this date (date or datetime objects or YYYY-MM-DD strings)
            latest_date: limit or finding tiles before this date
    '''

    _storage_cache = {}

    def __init__(
        self, 
        path: str, 
        aoi: Optional[Any] = None, 
        acq_id_in: Optional[Iterable[str]] = None,
        zone_in: Optional[Iterable[int]] = None, 
        earliest_date: Optional[Union[str, datetime, date]] = None,
        latest_date: Optional[Union[str, datetime, date]] = None, 
        profile: Optional[str] = None, 
        anon: Optional[bool] = False) -> None:
        ''' Initialize an ARDCollection from a path and optional filters

        Args:
            path: path to S3 prefix or STAC collection
            aoi: limit to finding tiles that cover this AOI, can be shapely geometry or most textual representations
            acq_id_in: limit to finding tiles from these acquisitions
            zone_in: limit to finding tiles in these zones
            earliest_date: limit to finding tiles after this date (date or datetime objects or YYYY-MM-DD strings)
            latest_date: limit or finding tiles before this date
            profile: AWS Profile to use when tiles are in S3
            anon: Access cloud data without authentication (for public buckets)
        '''

        self.acq_bins = None
    
        # TODO we may want to normalize AOIs to WGS84
        # but right now `covers` is probably capable enough
        self.aoi = aoi

        if anon:
            os.environ['AWS_NO_SIGN_REQUEST'] = 'YES'

        assert zone_in is None or all(type(z) == int for z in zone_in), 'Zones must be integers'
        if not zone_in:
            zone_in = None
        self.zone_in = zone_in 

        assert acq_id_in is None or all(type(c) == str for c in acq_id_in), 'Catalog IDs must be strings'
        if not acq_id_in:
            acq_id_in = None
        self.acq_id_in = acq_id_in 

        def format_date(d):
            if d is None:
                return None
            try:
                return d.strftime('%Y-%m-%d')
            except:
                assert re.match(r'\d{4}-\d{2}-\d{2}', d), "Dates must be YYYY-MM-DD strings, or date/datetime objects"
                return d

        self.earliest_date = format_date(earliest_date)
        '''limit to finding tiles after this date (date or datetime objects or YYYY-MM-DD strings)''' 
        self.latest_date = format_date(latest_date)
        '''limit to finding tiles before this date (date or datetime objects or YYYY-MM-DD strings)''' 

        self.path = path

        # this might need to be smarter (os.path?) for windows slashes
        if self.path[-1] == '/':
            self.path = self.path[:-1]
        if os.path.exists(self.path):
            self.storage_type = 'file'
            self.fs_path = os.path.abspath(self.path)
        elif self.path.startswith('s3://'):
            self.storage_type = 's3' 
            self.fs_path = self.path.replace('s3://', '')


        # cache a reference to the filesystem objects
        # if they are S3 they'll use file caching
        # we want the cached files to persist in the session
        if not self.path in ARDCollection._storage_cache:
            ARDCollection._storage_cache[self.path] = filesystem(self.storage_type, anon=anon, profile=profile)
        self.fs = ARDCollection._storage_cache[self.path]
        try:
            self.fs.ls(self.path)
        except:
            raise ValueError('Access error: check your path for errors and access permissions')

    def __repr__(self) -> str:
        return f'<ARDCollection at {self.path}/>'

    # This is an optimization, need to proxy through with the ARD API
    # instead of using GBDX
    # @lru_cache()
    # def get_footprint(self, acq_id):
    #     try:
    #         r = gbdx.get(catalog_url + acq_id)
    #         r.raise_for_status()
    #         footprint = (r.json()['properties']['footprintWkt'])
    #         return wkt.loads(footprint) 
    #     except Exception as e:
    #         warnings.warn(e)

    def _scan_files(self) -> None:
        self._dates = set()
        self._zones = set()
        # boundary optimization for acquisition ID searches
        #if self.acq_id_in is not None:
            # grab the bounds from the catalog to speed up fetching
            # uses just zones for now, could quadkey too
          #  if self.zone_in is None and self.aoi is None:
          #      lon2zone = lambda lon: (floor((lon + 180)/6) % 60) + 1
                #  for acq_id in self.acq_id_in:
                #     try:
                #         xmin, ymin, xmax, ymax = self.get_footprint(acq_id).bounds 
                #         print(xmin, ymin, xmax, ymax)
                #         self._zones.add(lon2zone(xmin))
                #         self._zones.add(lon2zone(xmax))
                #     except:
                #         pass 
        if self.zone_in is not None:
            for zone in self.zone_in:
                self._zones.add(zone)
        self.acq_bins = defaultdict(list)

        if self.aoi is not None:
            cells = set([f'{c.zone}/{c.quadkey}' for c in covers(self.aoi)])
        else:
            cells = []

        # STAC source
        if self.path.endswith('json'):

            # temp fix for paths
            # until we get stac links fixed
            def s(path):
                path = path.replace('ard-demo/sf-sample-02', 'maxar-ard-samples/sample-001')
                path = path.replace('s3://', '')
                return path

            with self.fs.open(self.path) as f:
                doc = json.load(f)

            # STAC Item
            if 'type' in doc.keys():
                items = [s(self.path)]

            # STAC Collection
            else:
                items = []
                for link in doc['links']:
                    if link['rel'] != 'child':
                        continue
                    with self.fs.open(s(link['href'])) as f:
                        links = json.load(f)['links']
                        for link in links:
                            if link['rel'] == 'item':
                                items.append(s(link['href']))

        # Filesystem source
        else:
            # build leading glob pattern based on zone & quadkey
            if self.aoi is not None:
                qkbs = set()
                for qk in cells:
                    qkbs.add(qk[:5])
                paths = [f'{k}*/*/*.json' for k in qkbs]
            else:
                if not self.zone_in or len(self.zone_in) == 0:
                    paths =['*/*/*/*.json']
                else: 
                    paths = [f'{z}/*/*/*.json' for z in self.zone_in]
            items = []
            for path in paths:
                full_path = f'{self.path}/{path}'
                for item in self.fs.glob(full_path):
                    items.append(item)

        # Filter out items
        for item in items:
            tile = ARDTile.from_doc(self.fs, item)
            if self.aoi is not None:
                if f'{tile.zone}/{tile.quadkey}' not in cells:
                    continue
            if self.acq_id_in is not None and tile.acq_id not in self.acq_id_in:
                continue
            if self.earliest_date is not None:
                if tile.date < self.earliest_date:
                    continue
            if self.latest_date is not None:
                if tile.date > self.latest_date:
                    continue
            self._dates.add(tile.date)
            self._zones.add(tile.cell.zone)
            self.acq_bins[tile.acq_id].append(tile)

    def clear_filesystem_cache(self):
        self.fs.clear_instance_cache()

    @classmethod
    def from_order(cls, order_id, **kwargs) -> ARDCollection:
        ''' Creates an collection from an order ID 

            Accepts all filtering parameters as used by class initialization.
        
        Returns: 
            ARDCollection '''

        if type(order_id) == Order:
            order = order_id
        else:
            order = Order.from_id(order_id)
        output = order.request.output_config
        path = f'{output.bucket}/{output.prefix}/order_collections/{order_id}_root_collection.json'
        self = cls(path, **kwargs)
        return self

    def _rescan(self):
        if self.acq_bins is None:
            self._scan_files()

    @property
    def tiles(self) -> List[ARDTile]:
        ''' All of the tiles in the ARDCollection '''
        self._rescan()
        return [tile for tiles in self.acq_bins.values() for tile in tiles]

    @property
    def acq_ids(self) -> List[str]:
        ''' All of the IDs of acquisitions in the ARDCollection '''
        self._rescan()
        return list(self.acq_bins.keys())

    @property
    def acquisitions(self) -> List[ARDTile]:
        ''' Acquisition objects of acquisitions in the ARDCollection '''
        self._rescan()
        return [Acquisition(acq_id, tiles) for acq_id, tiles in self.acq_bins.items()]

    @property
    def dates(self) -> List[str]:
        ''' All of the dates of acquisitions in the ARDCollection '''
        self._rescan()
        dates = list(self._dates)
        dates.sort()
        return dates

    @property
    def start_date(self) -> str:
        ''' Earliest date of acquisitions in the ARDCollection as YYYY-MM-DD '''
        self._rescan()
        return self.dates[0]

    @property
    def end_date(self) -> str:
        # does the type get caught?
        ''' Latest date of acquisitions in the ARDCollection as YYYY-MM-DD '''
        self._rescan()
        return self.dates[-1]

    @property
    def zones(self) -> List[int]:
        ''' The UTM zones covered by tiles in the ARDCollection '''
        self._rescan()
        return list(self._zones)

    def __setattr__(self, name, value):
        if name == 'acq_ids':
            raise ValueError('.acq_ids is read-only - set .acq_id_in instead')
        if name == 'zones':
            raise ValueError('.zones is read-only - set .zone_in instead')
        if name == 'start_date':
            raise ValueError('.start_date is read-only - set .earliest_date instead')
        if name == 'acq_ids':
            raise ValueError('.end_date is read-only - set .latest_date instead')
        if name in ['acq_id_in', 'zone_in', 'earliest_date', 'latest_date', 'aoi']:
            # resets the bins
            if self.acq_bins is not None:
                self.acq_bins = None
        object.__setattr__(self, name, value)




class Acquisition():
    ''' Container for one or more tiles from a single acquisition 
    
        These are generated by ARDCollections and would not normally be initialized '''

    def __init__(self, acq_id: str, tiles: List[ARDTile]) -> None:

        self.acq_id = acq_id
        ''' str: ID of this acquisition '''
        self.tiles = tiles
        ''' List[ARDTile]: List of ARDTiles in this acquisition '''

    def __repr__(self):
        return f'<Acquisition of {self.acq_id} ({len(self.tiles)} tiles)>'

    @property
    def properties(self) -> dict:
        ''' Strip level properties '''

        props = self.tiles[0].properties
        keep = ['catalog_id', 'platform', 'scan_direction','epsg','datetime']
        strip_props = {}
        for key in keep:
            strip_props[key] = props[key]
        return strip_props

            

def copy(collection: ARDCollection, destination: str, flatten: bool=False, dry_run: bool=False) -> None:
    ''' Copies data from a collection to a local location

    Args:
        collection: the ARDCollection to opy locally
        destination: local path to copy to
        flatten: write files in flat format without subdirectories
        dry_run: print the source and destination of files instead of copying '''
        
    Path(destination).mkdir(parents=True, exist_ok=True)
    for tile in collection.tiles:
        for path in tile.asset_paths.values():
            source = path
            target = path.replace(collection.fs_path, '')[1:]
            if flatten:
                target = urljoin(destination, target.replace('/','-'))
            else:
                target = urljoin(destination, target)
            if dry_run:
                print(source, target)
            else:
                if not flatten:
                    Path(target).parent.mkdir(parents=True, exist_ok=True)
                collection.fs.get(source, target)

